# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2019-02-19 01:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0026_data_scheduled"),
    ]

    operations = [
        migrations.AddField(
            model_name="data", name="purged", field=models.BooleanField(default=False),
        ),
    ]
